/*
 * Common utilities for filesystem users.
 *
 * Copyright (C) 2019, 2022 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <stdio.h>
#include <string.h>

#include "format.h"
#include "image.h"
#include "path.h"
#include "utils.h"



/* Show a directory listing for the given path. */

errcode_t utils_list_dir(ext2_filsys fs, const char *path)
{
    struct utils_list_dir_data data;

    data.fs = fs;

    /* Select a specific object, if appropriate. */

    if (!image_isdir(fs, path))
        data.filename = path_basename(path);
    else
        data.filename = NULL;

    /* Do not show link counts by default. */

    data.link_count = 0;

    return image_list_dir(fs, path, utils_list_dir_proc, &data);
}

/* Show directory entries when iterating. */

int utils_list_dir_proc(struct ext2_dir_entry *dirent, int offset,
                        int blocksize, char *buf, void *priv_data)
{
    struct utils_list_dir_data *data = (struct utils_list_dir_data *) priv_data;
    ext2_filsys fs = data->fs;
    struct ext2_inode inode;

    /* Select any indicated filename. */

    if ((data->filename != NULL) && (strcmp(dirent->name, data->filename)))
        return 0;

    /* Obtain the inode details for metadata. */

    if (ext2fs_read_inode(fs, dirent->inode, &inode))
        return DIRENT_ABORT;

    /* Output details in the style of "ls -l" showing directory, permissions,
       owner, group and size information. */

    printf("%s%s   %5d %5d      %6d ",
        _image_isdir(fs, dirent->inode) ? "d" : "-",
        get_permission_string(inode.i_mode),
        inode.i_uid,
        inode.i_gid,
        EXT2_I_SIZE(&inode));

    /* Show link counts if requested. */

    if (data->link_count)
        printf("%6d ", inode.i_links_count);

    /* Output the name which is presumably not necessarily null-terminated. */

    fwrite(dirent->name, sizeof(char), ext2fs_dirent_name_len(dirent), stdout);
    fputc((int) '\n', stdout);
    return 0;
}

/* vim: tabstop=4 expandtab shiftwidth=4
*/
